import inspect
import logging
from dataclasses import fields, field
from itertools import chain
from typing import Text, List

from .column import Column
from .compat import cdk, glue, s3
from .typemap import python_to_glue
from .util import operation

logger = logging.getLogger("monomer")


class monomer_property(property):

    def __init__(self, prp_or_column):
        self._column = None
        if isinstance(prp_or_column, Column):
            self._column = prp_or_column
        else:
            super().__init__(prp_or_column)

    def __call__(self, prp, *args, **kwargs):
        return super().__init__(prp)

    @staticmethod
    def persisted_properties(datacls):
        for k, v in inspect.getmembers(datacls, lambda a: isinstance(a, monomer_property)):
            yield v.make_field(k)

    def make_field(self, k):
        f = field()
        f.name = k
        _type = inspect.signature(self.fget).return_annotation
        if _type is inspect.Signature.empty:
            _type = Text
        f.type = _type
        if self._column is not None:
            f.metadata["monomer"] = {"column": self._column}
        return f


class MonomerDatabase(cdk.Construct):

    def __init__(self, scope: cdk.Construct, id: Text, database_name: Text):
        super().__init__(scope, id)
        self.db = glue.Database(self, 'lads', database_name='lads')
        self.s3bucket = s3.Bucket(self, f"{database_name}-store",
                                  bucket_name=f"{database_name}-data-storage",
                                  block_public_access=s3.BlockPublicAccess.BLOCK_ALL,
                                  encryption=s3.BucketEncryption.S3_MANAGED)
        self.tables: List[glue.Table] = []

    def add_table(self, datacls, bucket=None):
        dt_name = datacls.__name__
        columns = list(Column.from_field(dt_field)
                       for dt_field in chain(fields(datacls), monomer_property.persisted_properties(datacls)))
        table = glue.Table(self, f'{dt_name}Table',
                           database=self.db,
                           table_name=dt_name,
                           bucket=bucket or self.s3bucket,
                           s3_prefix=dt_name.lower() if bucket is None else '',
                           columns=list(col.to_glue() for col in columns),
                           data_format=glue.DataFormat.JSON)
        self.tables.append(table)
        return table



from dataclasses import dataclass, field, Field
from typing import Text

from .compat import glue, Type
from .typemap import python_to_glue


def extract_meta(dt_field: Field):
    dt_meta = dt_field.metadata.copy()
    _meta = {}
    if isinstance(dt_meta, Column):
        _meta["column"] = dt_meta
    else:
        _meta = dt_meta.pop("monomer", _meta)
        for k in dt_meta:
            if k[:7] == "_monomer":
                _meta[k[7:]] = dt_meta[k]
    return _meta


@dataclass
class Column(object):
    name: Text
    type: Type = field(default=None)
    comment: Text = field(default="Monomer generated column")

    def to_glue(self):
        return glue.Column(name=self.name, type=self.type, comment=self.comment)

    @classmethod
    def from_field(cls, dt_field):
        meta = extract_meta(dt_field)
        column = meta.get("column", Column(dt_field.name))
        if column.name is None:
            column.name = dt_field.name
        if column.type is None:
            column.type = python_to_glue(dt_field.type)
        return column



import datetime
from dataclasses import dataclass, field
from typing import Union, Text
from unittest import mock


try:
    import aws_cdk.core as cdk
    import aws_cdk.aws_s3 as s3
    import aws_cdk.aws_glue as glue
    import aws_cdk.aws_athena as athena
    import jsii
    Type = Union[type, glue.Type]

except ImportError:
    cdk = mock.Mock()
    s3 = mock.Mock()
    glue = mock.Mock()
    athena = mock.Mock()
    Type = type

Datetime = datetime.datetime
Date = datetime.date


def array(item_type: glue.Type) -> glue.Type:
    return jsii.sinvoke(glue.Schema, "array", [item_type])


def glue_map(key_type, value_type):
    return jsii.sinvoke(glue.Schema, "map", [key_type, value_type])



from dataclasses import asdict

import datetime
from functools import wraps



Datetime = datetime.datetime


class _Empty: ...

def operation(func):
    @wraps(func)
    def wrapper(instance, item, session=None):
        instance.ensure_table_exists(item)

        commit = session is None
        session = session
        ret = func(instance, item, session)



from typing import Type, Union, Text, List, get_args, get_origin, Any
from .util import Datetime
from dataclasses import dataclass, is_dataclass, fields, Field, field
from .compat import glue, Datetime, Date, array, glue_map

VARCHAR_LEN = 256

glue_type_map = {
    Text: glue.Schema.STRING,
    Datetime: glue.Schema.TIMESTAMP,
    float: glue.Schema.FLOAT,
    int: glue.Schema.INTEGER,
    Date: glue.Schema.DATE
}




def python_to_glue(a_type):
    try:
        if issubclass(a_type, glue.Type):
            return a_type
    except TypeError:
        ...

    generic_type = get_origin(a_type)
    if generic_type:
        template = list(get_args(a_type))
        if generic_type == dict:
            return glue_map(python_to_glue(template[0]), python_to_glue(template[1]))
        if generic_type == list:
            return array(python_to_glue(template[0]))
        if generic_type == Union:
            try:
                template.remove(type(None))
            except ValueError:
                ...  # no worries if it isn't in there.
        if len(template) == 1:
            return python_to_glue(template[0])

    return glue_type_map.get(a_type, glue.Schema.STRING)
